/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    alo0ng with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Eclat.java
 */

package weka.associations;

import java.util.Enumeration;

import weka.core.Capabilities;
import weka.core.FastVector;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.RevisionUtils;
import weka.core.TechnicalInformation;
import weka.core.TechnicalInformationHandler;
import weka.core.Utils;
import weka.core.Capabilities.Capability;
import weka.core.TechnicalInformation.Field;
import weka.core.TechnicalInformation.Type;


public class Eclat 
  extends AbstractAssociator 
  implements OptionHandler, CARuleMiner, TechnicalInformationHandler {
  
  /** for serialization */
  static final long serialVersionUID = 3277498842319212687L;
  
  /** The minimum support. */
  protected double m_minSupport;
  
  /** The maximum number of rules that are output. */
  protected int m_numRules;

  /** The list of all generated rules. */
  protected FastVector[] m_largeItemsets;

  /** The instances (transactions) to be used for generating 
      the association rules. */
  protected Instances m_instances;

  /** Only the class attribute of all Instances.*/
  protected Instances m_onlyClass;
  
  /** The class index. */
  protected int m_classIndex;
  
  
  /**
   * Constructor that allows to sets default values for the 
   * minimum confidence and the maximum number of rules
   * the minimum confidence.
   */
  public Eclat() {
    resetOptions();
  }

  /**
   * Resets the options to the default values.
   */
  public void resetOptions() {
    m_numRules = 10;
    m_minSupport = 0.1;
    m_classIndex = -1;
  }

  /**
   * Returns default capabilities of the classifier.
   *
   * @return      the capabilities of this classifier
   */
  public Capabilities getCapabilities() {
    Capabilities result = super.getCapabilities();

    // attributes
    result.enable(Capability.BINARY_ATTRIBUTES);
    result.enable(Capability.NO_CLASS);

    return result;
  }
  
  public TechnicalInformation getTechnicalInformation() {
    TechnicalInformation 	result;
    TechnicalInformation 	additional;
    
    result = new TechnicalInformation(Type.INPROCEEDINGS);
    result.setValue(Field.AUTHOR, "R. Agrawal and R. Srikant");
    result.setValue(Field.TITLE, "Fast Algorithms for Mining Association Rules in Large Databases");
    result.setValue(Field.BOOKTITLE, "20th International Conference on Very Large Data Bases");
    result.setValue(Field.YEAR, "1994");
    result.setValue(Field.PAGES, "478-499");
    result.setValue(Field.PUBLISHER, "Morgan Kaufmann, Los Altos, CA");

    additional = result.add(Type.INPROCEEDINGS);
    additional.setValue(Field.AUTHOR, "Bing Liu and Wynne Hsu and Yiming Ma");
    additional.setValue(Field.TITLE, "Integrating Classification and Association Rule Mining");
    additional.setValue(Field.BOOKTITLE, "Fourth International Conference on Knowledge Discovery and Data Mining");
    additional.setValue(Field.YEAR, "1998");
    additional.setValue(Field.PAGES, "80-86");
    additional.setValue(Field.PUBLISHER, "AAAI Press");
    
    return result;
  }

  /**
   * Method that generates all large itemsets with a minimum support, and from
   * these all association rules with a minimum confidence.
   *
   * @param instances the instances to be used for generating the associations
   * @throws Exception if rules can't be built successfully
   * @throws IllegalMinimalSupportException 
   */
  public void buildAssociations(Instances instances) throws Exception {
    instances.setClassIndex(instances.numAttributes()-1);
    getCapabilities().testWithFail(instances);
    m_instances = instances;
    EclatDatabase eclat = new EclatDatabase(instances, this.m_minSupport);
    this.m_largeItemsets = eclat.callEclat();
    System.out.println(toString());
  }
  
  
      /**
     * Method that mines all class association rules with minimum support and
     * with a minimum confidence.
     * @return an sorted array of FastVector (confidence depended) containing the rules and metric information
     * @param data the instances for which class association rules should be mined
     * @throws Exception if rules can't be built successfully
     */
    public FastVector[] mineCARs(Instances data) throws Exception {
	buildAssociations(data);
	return m_largeItemsets;
    }

   /**
   * Gets the instances without the class atrribute.
   *
   * @return the instances without the class attribute.
   */ 
  public Instances getInstancesNoClass() {
      
      return m_instances;
  }  
  
  
  /**
   * Gets only the class attribute of the instances.
   *
   * @return the class attribute of all instances.
   */ 
  public Instances getInstancesOnlyClass() {
      
      return m_onlyClass;
  }  


  /**
   * Returns an enumeration describing the available options.
   *
   * @return an enumeration of all the available options.
   */
  public Enumeration listOptions() {

    String string1 = "\tThe required number of rules. (default = " + m_numRules + ")",
      string2 = "\tThe minimum support for the itemsets\n";
    

    FastVector newVector = new FastVector(2);

    newVector.addElement(new Option(string1, "N", 1, 
				    "-N <required number of rules output>"));
    newVector.addElement(new Option(string2 + "\tMinimum support "
				    +"(default = 0.2)", "M", 1,
				     "-M <for minimum support>"));
    
    return newVector.elements();
  }

  /**
   * Parses a given list of options. <p/>
   * 
   <!-- options-start -->
   * Valid options are: <p/>
   * 
   * * <pre> -N &lt;minimum support&gt;
   *  The minimum support for an itemset. (default = 0.1)</pre>
   *  
   * <pre> -N &lt;required number of rules output&gt;
   *  The required number of rules. (default = 10)</pre>
   *  
   <!-- options-end -->
   *
   * @param options the list of options as an array of strings
   * @throws Exception if an option is not supported 
   */
  public void setOptions(String[] options) throws Exception {
    
    resetOptions();
    String numRulesString = Utils.getOption('N', options),
    minSuppString = Utils.getOption('M', options);
    
    if (numRulesString.length() != 0) {
      m_numRules = Integer.parseInt(numRulesString);
    }
    if (minSuppString.length() != 0) {
      this.m_minSupport = (new Double(minSuppString)).doubleValue();
    }
  }

  /**
   * Gets the current settings of the Apriori object.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    String [] options = new String [20];
    int current = 0;

    options[current++] = "-N"; options[current++] = "" + m_numRules;
    options[current++] = "-M"; options[current++] = "" + m_minSupport;
    
    while (current < options.length) {
      options[current++] = "";
    }
    return options;
  }

  /**
   * Outputs the size of all the generated sets of itemsets and the rules.
   * 
   * @return a string representation of the model
   */
  public String toString() {

    StringBuffer text = new StringBuffer();

    text.append("\nEclat\n=======\n\n");
    text.append("Minimum support: " 
		+ Utils.doubleToString(m_minSupport,2) 
		+ '\n');
   
    text.append("\nGenerated sets of large itemsets:\n");
    
    int max = m_numRules;
    if (m_largeItemsets.length < m_numRules) {
      max = m_largeItemsets.length;
    }
    
    for (int i = 0; i < max; i++) {
        FastVector rule = m_largeItemsets[i];
        for (int j = 0; j < rule.size(); j++) {
          text.append(this.m_instances.attribute((Integer) rule.elementAt(j)).name() + " ");
        }
        text.append("\n");
    }
    return text.toString();
  }
  
   /**
   * Returns the metric string for the chosen metric type
   * @return a string describing the used metric for the interestingness of a class association rule
   */
  public String metricString() {
      return "metric string";
  }

   /**
   * Sets the class index
   * @param index the class index
   */  
  public void setClassIndex(int index){
      
      m_classIndex = index;
  }
  
  /**
   * Returns the tip text for this property
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui
   */
  public String numRulesTipText() {
    return "Number of rules to find.";
  }

  /**
   * Get the value of numRules.
   *
   * @return Value of numRules.
   */
  public int getNumRules() {
    
    return m_numRules;
  }
  
  /**
   * Set the value of numRules.
   *
   * @param v  Value to assign to numRules.
   */
  public void setNumRules(int v) {
    
    m_numRules = v;
  }
  
  /**
   * Returns the tip text for this property
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui
   */
  public String minSupportTipText() {
    return "The Minimum Support for an itemset.";
  }

  /**
   * Get the value of minSupport.
   *
   * @return Value of minSupport.
   */
  public double getMinSupport() {
    
    return m_minSupport;
  }
  
  /**
   * Set the value of minSupport.
   *
   * @param v  Value to assign to minSupport.
   */
  public void setMinSupport(double v) {
    
    m_minSupport = v;
  }

  /**
   * Returns the revision string.
   * 
   * @return		the revision
   */
  public String getRevision() {
    return RevisionUtils.extract("$Revision: 1.0 $");
  }

  /**
   * Main method.
   * 
   * @param args the commandline options
   */
  public static void main(String[] args) {
    runAssociator(new Eclat(), args);
  }
}

