package weka.associations;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

import weka.core.FastVector;
import weka.core.Instances;

public class EclatDatabase {

  private Map<Integer, Set<Integer>> tidlists;
  Set<Integer> itemset;
  private int minSupport;
  
  /**
   * Constructor based on a 0/1-database
   * - make a tidlist for every item
   * - filter out the infrequent items
   * @param data the data
   * @param minSupport the minimal support (percentage)
   * @throws IllegalMinimalSupportException should be between 0 and 1
   * @throws Exception 
   */
  public EclatDatabase(Instances instances, double minSupport) throws Exception {
    if ((minSupport <= 0) || (minSupport > 1)) { // the minimal support should be higher than 0 and not higher than 1
      throw new Exception("Illegal minimal support: "  + minSupport);
    }
    this.minSupport = (int) Math.ceil(instances.numInstances() * minSupport);
    
    this.itemset = new TreeSet<Integer> ();
    
    // create a tidlist for every item
    this.tidlists = new TreeMap<Integer, Set<Integer>> ();
    for (int itemNr = 0; itemNr < instances.numAttributes(); itemNr++) {
      Set<Integer> tidlist = new TreeSet<Integer>();
      this.tidlists.put(itemNr, tidlist);
    }
    
    // fill up the right tidlist when a 1 is encountered
    for (int transaction = 0; transaction < instances.numInstances(); transaction++) {
      for (int itemNr = 0; itemNr < instances.numAttributes(); itemNr++) {
	if (instances.instance(transaction).value(itemNr) > 0) {
	  this.tidlists.get(itemNr).add(transaction);
	}
      }
    }
      
    // we only need the frequent items
    for (int itemNr = this.tidlists.size() - 1; itemNr >= 0; itemNr--) {
      int support = this.tidlists.get(itemNr).size();
      if (support < this.minSupport) {
	this.tidlists.remove(itemNr);
      }
    }
  }
  
  /**
   * Constructor used by the algorithm to create a conditional database
   * @param tidlists
   * @param itemset
   * @param minSupport
   */
  private EclatDatabase (Map<Integer, Set<Integer>> tidlists, Set<Integer> itemset, int minSupport) {
    this.tidlists = tidlists;
    this.itemset = itemset;
    this.minSupport = minSupport;
  }
  
  /**
   * Getter for the tidlists
   * @return the tidlists in the database
   */
  public Map<Integer, Set<Integer>> getTIdLists () {
    return this.tidlists;
  }
  
  /**
   * Getter for the itemset
   * @return the itemset
   */
  public Set<Integer> getItemSet () {
    return this.itemset;
  }
  
  public FastVector[] callEclat() {
    Vector<Set<Integer>> result = this.Eclat();
    FastVector[] copy = new FastVector[result.size()];
    for (int i = 0; i < result.size(); i++) {
      copy[i] = Utility.SetToFastVector(result.get(i));
    }
    return copy;
  }
  
  /**
   * Calls the Eclat Algorithm
   * @return the frequent itemsets
   */
  private Vector<Set<Integer>> Eclat () {
    Vector<Set<Integer>> itemsets_I = new Vector<Set<Integer>> ();
    
    Iterator<Integer> keys1 = this.tidlists.keySet().iterator();
    while (keys1.hasNext()) { // loop for every item in the database
      Integer i = keys1.next();
      //Create D^i: conditional database on i

      EclatDatabase conditional_i = this.conditional_database(i);
      itemsets_I.add(conditional_i.itemset);
    
      // if conditional database not empty, apply recursively
      if (conditional_i.tidlists.size() > 0) {
	Vector<Set<Integer>> itemsets_I_i = conditional_i.Eclat();
	Utility.addList(itemsets_I, itemsets_I_i);
      }
    }
    return itemsets_I;
  }
  
  /**
   * Creates a conditional database over a given item
   * @param i the item over which we want to create a conditional database
   * @return the conditional database
   */
  public EclatDatabase conditional_database (Integer i) {
    Set<Integer> tidlist_i = this.tidlists.get(i);
    Map<Integer, Set<Integer>> conditional_i = new TreeMap<Integer, Set<Integer>> ();
    
    Set<Integer> copy = new TreeSet<Integer> (this.tidlists.keySet());
    Iterator<Integer> keys2 = copy.iterator();
    
    Integer j = keys2.next();
    while (j != i) { // loop till (j == i)
      j = keys2.next();
    }
    while (keys2.hasNext()) { // check for every j > i whether the intersection of their tidlists is frequent
      j = keys2.next();
      Set<Integer> intersection = Utility.intersectSets(tidlist_i, this.tidlists.get(j));
      if (intersection.size() >= this.minSupport) {
	conditional_i.put(j, intersection);
      }
    }
    
    Set<Integer> itemset_i = new TreeSet<Integer> (this.itemset);
    itemset_i.add(i); // add the item to the current itemset, we know the result will be frequent
    return new EclatDatabase(conditional_i, itemset_i, this.minSupport);
  }
}
