package weka.associations;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

import weka.associations.Utility.DBTidlists;
import weka.core.FastVector;
import weka.core.Instances;

public class LTMDatabase {
  
  protected Map<Integer, Set<Integer>> tidlists;
  protected Map<Integer, Integer> sizes;
  protected Set<Integer> itemset;
  
  /**
   * Constructor used by the algorithm to create a conditional database
   * @param tidlists for every item we need a tidlist
   * @param sizes for every item we need the size
   * @param conditional_on the database is conditional over this item set
   */
  protected LTMDatabase(Map<Integer, Set<Integer>> tidlists, Map<Integer, Integer> sizes, Set<Integer> conditional_on) {
    this.tidlists = tidlists;
    this.sizes = sizes;
    this.itemset = conditional_on;
  }
  
  /**
   * Constructor based on a 0/1-database
   * - make a tidlist for every item
   * - calculate the size (number of 1's occuring) for every transaction
   * @param data
   */
  public LTMDatabase(Instances instances) {
    DBTidlists db = Utility.makeTidlists(instances);
    this.sizes = db.sizes;
    this.tidlists = db.tidlists;
    
    this.itemset = new TreeSet<Integer> ();
  }
  
  /**
   * Getter for the tidlists
   * @return the tidlists
   */
  public Map<Integer, Set<Integer>> getTIdLists () {
    return this.tidlists;
  }
  
  /**
   * Getter for the sizes
   * @return the sizes
   */
  public Map<Integer, Integer> getSizes () {
    return this.sizes;
  }
  
  /**
   * Getter for the itemset
   * @return the itemset
   */
  public Set<Integer> getItemSet () {
    return this.itemset;
  }
  
  /**
   * Removes an item from the database
   * Herefore the size of all transactions where the item occurs must be adapted
   * @param itemNr number of the item that has to be removed
   */
  public void removeItem (Integer itemNr) {
    Iterator<Integer> tidlist = this.tidlists.get(itemNr).iterator();
    while (tidlist.hasNext()) {
      this.lowerSize(tidlist.next()); // lower size of every transaction where the item occurs
    }
    this.tidlists.remove(itemNr); // remove the item
  }
  
  /**
   * Lowers the size of a given transaction by 1
   * @param tid the id of the transaction
   */
  private void lowerSize (Integer tid) {
    int new_size = this.sizes.get(tid) - 1;
    this.sizes.remove(tid);
    if (new_size > 0) { // no need to keep it if it will never occur in a tidlist
      this.sizes.put(tid, new_size);
    }
  }
  
  /**
   * Create D^i: conditional database over i
   * @param i one of the items
   * @return the tidlists of the conditional database
   */
  protected Map<Integer, Set<Integer>> conditional_tidlists (Integer i) {
    Map<Integer, Set<Integer>> conditional_i = new TreeMap<Integer, Set<Integer>> ();
    Set<Integer> tidlist_i = this.tidlists.get(i);
    
    Set<Integer> copy = new TreeSet<Integer> (this.tidlists.keySet());
    Iterator<Integer> keys = copy.iterator();
    
    Integer j = keys.next();
    while (j != i) { // loop till (j == i)
      j = keys.next();
    }
    // for every j > i, make the intersection of the tidlists of i and j, then add it to the database
    while (keys.hasNext()) {
      j = keys.next();
      Set<Integer> intersection = Utility.intersectSets(tidlist_i, this.tidlists.get(j));
      conditional_i.put(j, intersection);
    }
    
    return conditional_i;
  }
  
  /**
   * counts the occurrence of the sizes of the transactions for a given item 
   * @param amounts the amounts are stored here
   * @param itemNr the number of the item
   * @return the maximum size of the transactions
   */
  public int countTransactionSizes (Map<Integer, Integer> counts, Integer itemNr) {
    Iterator<Integer> transactions = this.tidlists.get(itemNr).iterator();
    int maximum = 0;
    
    // loop for every transaction
    while (transactions.hasNext()) {
      Integer transaction = transactions.next();
      int size = this.sizes.get(transaction);
      if (counts.containsKey(size)) { // increase the counter for this size
				int amount = counts.get(size) + 1;
      	counts.remove(size);
      	counts.put(size, amount);
      }
      else { // add an item for this size with count 1
				counts.put(size, 1);
      }
      
      if (size > maximum) {
      	maximum = size;
      }
    }
    return maximum;
  }
  
  /**
   * Creates an array with the sizes of all transactions
   * The current size are used, and are decreased by 1 for every transaction of a given item
   * @param itemNr
   * @return
   */
  protected Map<Integer, Integer> createSizeArray(Integer itemNr) {
    Map<Integer, Integer> new_size = new TreeMap<Integer, Integer> ();
    
    Set<Integer> tidlist = this.tidlists.get(itemNr);
    
    Iterator<Integer> transactions = this.sizes.keySet().iterator();
    while (transactions.hasNext()) {
      Integer transaction = transactions.next();
      Integer size = this.sizes.get(transaction); // get the current size
      
      if (tidlist.contains(transaction)) { // decrease by 1
	size--;
      }
      if (size > 0) { // no need to keep a size for a transaction that doesn't occur in a tidlist
      	new_size.put(transaction, size);
      }
    }
    
    return new_size;
  }
  
  /**
   * Create a database that is conditional over a given item
   * @param itemNr
   * @return the conditional database
   */
  public LTMDatabase conditional_database (Integer itemNr) {
    Map<Integer, Set<Integer>> new_tidlists = this.conditional_tidlists(itemNr);
    if (new_tidlists.size() > 0) {
      Map<Integer, Integer> new_sizes = this.createSizeArray(itemNr);
      Set<Integer> new_itemset = new TreeSet<Integer> (this.itemset);
      new_itemset.add(itemNr);
      LTMDatabase new_db = new LTMDatabase(new_tidlists, new_sizes, new_itemset);
      return new_db;
    }
    else {
      return null;
    }
  }
  
  /**
   * Removes a transaction from the tidlist of a given item
   * @param itemNr
   * @param tid
   */
  public void removeTransactionFromTidlist (Integer itemNr, Integer tid) {
    this.tidlists.get(itemNr).remove(tid);
    this.lowerSize(tid);
  }
  
  public FastVector[] callLTM(int minSupport) {
    Vector<Set<Integer>> result = this.LTM(minSupport);
    FastVector[] copy = new FastVector[result.size()];
    for (int i = 0; i < result.size(); i++) {
      copy[i] = Utility.SetToFastVector(result.get(i));
    }
    return copy;
  }
  
  /**
   * Calls the LTM algorithm
   * @param minTiling the minimum size of a large tile
   * @return the large tiles
   */
  private Vector<Set<Integer>> LTM (int minTiling) {
    Vector<Set<Integer>> itemsets_I = new Vector<Set<Integer>> (); // start with empty set
    
    // prune the database
    this.prune(minTiling);
    
    Iterator<Integer> keys = this.tidlists.keySet().iterator();
    while (keys.hasNext()) { // loop for every item in the database
      Integer i = keys.next();
      
      // If a large tile, add the set to the set of large tiles
      if (this.tidlists.get(i).size() * (this.itemset.size() + 1) >= minTiling) {
      	// copy the itemset and add i
      	Set<Integer> itemset_i = new TreeSet<Integer> (this.itemset);
      	itemset_i.add(i);
      	itemsets_I.add(itemset_i);
      }
      
      //Create D^i: conditional database over i
      LTMDatabase conditional_i = this.conditional_database(i);
      
      // if conditional database not empty, apply recursively
      if (conditional_i != null) {
      	Vector<Set<Integer>> itemsets_I_i = conditional_i.LTM(minTiling);
      	Utility.addList(itemsets_I, itemsets_I_i); // add all the itemsets
      }
    }
    return itemsets_I;
  }
  
  
  /**
   * Calls the top-k algorithm for large tiles
   * @param k
   * @return the k largest tiles of the database
   */
  public FastVector[] CallTopK (int k) {
    Vector<Integer> sizes = new Vector<Integer> (k);
    Vector<Set<Integer>> result = this.topK(k, sizes);
    FastVector[] copy = new FastVector[result.size()+1];
    
    FastVector copySizes = new FastVector(sizes.size());
    for (int i = 0; i < k; i++) {
      copySizes.addElement(sizes.get(i));
    }
    
    copy[0] = copySizes;
    
    for (int i = 0; i < result.size(); i++) {
      copy[i+1] = Utility.SetToFastVector(result.get(i));
    }
    return copy;
  }
  
  /**
   * Calls the top-k algorithm for large tiles
   * @param k
   * @return the k largest tiles of the database
   */
  public Vector<Set<Integer>> topK (int k, Vector<Integer> sizes) {
    Vector<Set<Integer>> tiles = new Vector<Set<Integer>> (k);
    tiles.add(new TreeSet<Integer> ());
    sizes.add(0);
    this.LTMTopK(tiles, sizes);
    
    return tiles;
  }
  
  /**
   * Calls the LTM algorithm 
   * @param tiles list of tiles found so far
   * @param sizes list of sizes for the tiles found so far
   */
  public void LTMTopK (Vector<Set<Integer>> tiles, Vector<Integer> sizes) {
    // prune the database
    this.prune(sizes.get(0));
    
    Iterator<Integer> keys = this.tidlists.keySet().iterator();
    while (keys.hasNext()) { // loop for every item in the database
      
      Integer i = keys.next();
      
      int my_size = this.tidlists.get(i).size() * (this.itemset.size() + 1);
      
      // If a large tile, add the set to the set of large tiles
      if (my_size > sizes.get(0)) {
      	// copy the itemset and add i
      	Set<Integer> itemset_i = new TreeSet<Integer> (this.itemset);
      	itemset_i.add(i);
      	
      	this.insertTile(tiles, sizes, itemset_i, my_size);
      	
      }
      
      //Create D^i: conditional database over i
      LTMDatabase conditional_i = this.conditional_database(i);
      
      // if conditional database not empty, apply recursively
      if (conditional_i != null) {
      	conditional_i.LTMTopK(tiles, sizes);
      }
    }
  }
  
  public void insertTile (Vector<Set<Integer>> tiles, Vector<Integer> sizes
      				, Set<Integer> my_tile, int my_size) {
    if (tiles.size() == tiles.capacity()) {
      tiles.remove(0);
      sizes.remove(0);
    }
    for (int index = 0; index < tiles.size(); index++) {
      if (my_size <= sizes.get(index)) {
      	tiles.insertElementAt(my_tile, index);
      	sizes.insertElementAt(my_size, index);
      	return;
      }
    }
    tiles.add(my_tile);
    sizes.add(my_size);
  }
  
  /**
   * Function to prune the database
   * @param minTiling the minimum size of a large tile
   */
  public void prune (int minTiling) {
    boolean changed; // check whether something has changed while pruning
    do {
      changed = false;
      Set<Integer> copy = new TreeSet<Integer> (this.tidlists.keySet());
      Iterator<Integer> keys = copy.iterator();
      while (keys.hasNext()) { // loop for every item in the database
      	Integer i = keys.next();
      	
      	Map<Integer, Integer> counts = new TreeMap<Integer, Integer> ();
      	int maximum = this.countTransactionSizes(counts, i);
      	
      	// check whether it's possible to create a large tile with this item
      	boolean possible = false;
      	int amount = 0; // amount counts all the transactions with a size which is >= than the current size
      	int ML = maximum;
      	// start with the highest sizes, then count down to the lowest
      	for (int size = maximum; size > 0; size--) {
          if (counts.containsKey(size)) {
            amount += counts.get(size);
            // the maximum area that can be covered with this amount and size
            int upper_bound = amount * (this.itemset.size() + size);
            if (upper_bound >= minTiling) {
              possible = true;
              ML = size; // we want to keep the minimum size for which a large tile is still possible
            }
          }
      	}
      	if (!possible) { // if not possible, remove the item
      	  changed = true;
      	  this.removeItem(i);
      	}
      	else { // it's possible to create a large tile, but maybe not for all transactions
      		
      	  Set<Integer> tidlist_i = this.tidlists.get(i);
      	  copy = new TreeSet<Integer> (tidlist_i);
      	  Iterator<Integer> transactions = copy.iterator();
      	
      	  while (transactions.hasNext()) {
      	    Integer transaction = transactions.next();
      	    // remove if size < minimal size needed
      	    if (this.sizes.get(transaction) < ML) {
      	      this.removeTransactionFromTidlist(i, transaction);
      	      changed = true;
      	    }
      	  }
      	}
      }
    } while (changed); // loop till nothing changed
  }
}
