package weka.associations;

import java.sql.Time;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

import weka.core.FastVector;
import weka.core.Instances;

public class TilingOverlapDatabase extends LTMDatabase {
  
  private Map<Integer, Set<Integer>> uncovered_tidlists;
  private Map<Integer, Integer> uncovered_sizes;
  private Map<Integer, Set<Integer>> uncovered_items;
	private Map<Integer, Set<Integer>> items_still_to_cover;
  
  /**
   * Constructor used by the algorithm to create a conditional database
   * @param tidlists for every item we need a tidlist
   * @param sizes for every item we need the size
   * @param conditional_on the database is conditional over this item set
   */
  public TilingOverlapDatabase(Map<Integer, Set<Integer>> tidlists, Map<Integer, Integer> sizes
      				, Set<Integer> conditional_on, Map<Integer, Set<Integer>> uncovered_tidlists
      				, Map<Integer, Integer> uncovered_sizes, Map<Integer, Set<Integer>> uncovered_items) {
    super(tidlists, sizes, conditional_on);
    this.uncovered_tidlists = uncovered_tidlists;
    this.uncovered_sizes = uncovered_sizes;
    this.uncovered_items = uncovered_items;
  }
  
  /**
   * Constructor used by the algorithm to create a conditional database
   * @param tidlists for every item we need a tidlist
   * @param sizes for every item we need the size
   * @param conditional_on the database is conditional over this item set
   */
  public TilingOverlapDatabase(Map<Integer, Set<Integer>> tidlists, Map<Integer, Integer> sizes, Set<Integer> conditional_on) {
    super(tidlists, sizes, conditional_on);
    this.initOverlaps();
  }
  
  /**
   * Constructor based on a 0/1-database
   * @param data
   */
  public TilingOverlapDatabase(Instances instances) {
    super(instances);
    this.initOverlaps();

		this.items_still_to_cover = new TreeMap<Integer, Set<Integer>> ();
		for (int transaction = 0; transaction < instances.numInstances(); transaction++) {
			Set<Integer> items = new TreeSet<Integer>();
			for (int itemNr = 0; itemNr < instances.numAttributes(); itemNr++) {
				if (instances.instance(transaction).value(itemNr) > 0) {
					items.add(itemNr);
				}
			}
			this.items_still_to_cover.put(transaction, items);
		}
  }
  
  public void initOverlaps() {
    this.uncovered_tidlists = new TreeMap<Integer, Set<Integer>> ();
    Iterator<Integer> it = this.tidlists.keySet().iterator();
    while (it.hasNext()) {
      int key = it.next();
      this.uncovered_tidlists.put(key, new TreeSet<Integer> (this.tidlists.get(key)));
    }
    
    this.uncovered_sizes = new TreeMap<Integer, Integer>();
    it = this.sizes.keySet().iterator();
    while (it.hasNext()) {
      int key = it.next();
      this.uncovered_sizes.put(key, this.sizes.get(key));
    }
    
    this.uncovered_items = new TreeMap<Integer, Set<Integer>> ();
  }
  
  public void DeleteTile(Set<Integer> tile) {
//		System.out.println("DEBUG 0: this.sizes = " + this.sizes);
//		System.out.println("DEBUG 0: this.uncovered_sizes = " + this.uncovered_sizes);
//		System.out.println("DEBUG: tile = " + tile);
    //Get the intersection of the tidlists of the items in the tile
    Iterator<Integer> it = tile.iterator();
    Set<Integer> intersection = new TreeSet<Integer> (this.tidlists.get(it.next()));
    while (it.hasNext()) {
      Set<Integer> tidlist = new TreeSet<Integer> (this.tidlists.get(it.next()));
      intersection = Utility.intersectSets(intersection, tidlist);
    }
//   	System.out.println("DEBUG: intersection = " + intersection); 

    it = tile.iterator();
    Set<Integer> copy = new TreeSet<Integer> (intersection);
    while (it.hasNext()) {
      int key = it.next();
      Utility.diffSets(this.uncovered_tidlists.get(key), copy);
    }
    
    it = intersection.iterator();
    while (it.hasNext()) {
      int key = it.next();
      Integer size = this.uncovered_sizes.get(key);
      this.uncovered_sizes.remove(key);
//      this.uncovered_sizes.put(key, size - tile.size()); // BUG
			Utility.diffSets(this.items_still_to_cover.get(key), tile);
			this.uncovered_sizes.put(key, this.items_still_to_cover.get(key).size());	
    }
//		System.out.println("DEBUG 1: this.sizes = " + this.sizes);
//		System.out.println("DEBUG 1: this.uncovered_sizes = " + this.uncovered_sizes);
  }
  
  /**
   * Removes an item from the database
   * Herefore the size of all transactions where the item occurs must be adapted
   * @param itemNr number of the item that has to be removed
   */
  public void removeItem (Integer itemNr) {
    Iterator<Integer> tidlist = this.tidlists.get(itemNr).iterator();
    while (tidlist.hasNext()) {
      this.lowerSize(tidlist.next()); // lower size of every transaction where the item occurs
    }
    this.tidlists.remove(itemNr); // remove the item
    this.uncovered_tidlists.remove(itemNr); // remove the item
  }
 
	int debug_size = -1;
	int debug_uncovered_size = -1; 
  /**
   * Lowers the size of a given transaction by 1
   * @param tid the id of the transaction
   */
  private void lowerSize (Integer tid) {
try {
//		if (tid == 7) {
//			System.out.println("DEBUG: this.sizes = " + this.sizes);
//			System.out.println("DEBUG: this.uncovered_sizes = " + this.uncovered_sizes);
//		}
		this.debug_size = -1;
    int new_size = this.sizes.get(tid) - 1;
		this.debug_size = new_size;
    this.sizes.remove(tid);
		this.debug_uncovered_size = -1;
    int new_uncovered_size = this.uncovered_sizes.get(tid)/* - 1*/; // BUG: null pointer exception
/*
		int new_uncovered_size = 0;
		if (this.uncovered_sizes.get(tid) != null) {
			new_uncovered_size = this.uncovered_sizes.get(tid) - 1;
		}
*/
		assert new_uncovered_size >= 0;
		this.debug_uncovered_size = new_uncovered_size;
    this.uncovered_sizes.remove(tid);

//    if (new_size > 0) { // no need to keep it if it will never occur in a tidlist
      this.sizes.put(tid, new_size);
      this.uncovered_sizes.put(tid, new_uncovered_size);
//    }
/*
		if (new_uncovered_size > 0) {
			this.uncovered_sizes.put(tid, new_uncovered_size);	
		}
*/
} catch (Exception e) {
	System.out.println("DEBUG: tid = " + tid);
	System.out.println("DEBUG: new_size = " + this.debug_size);
	System.out.println("DEBUG: this.sizes = " + this.sizes);
	System.out.println("DEBUG: new_uncovered_size = " + this.debug_uncovered_size);
	System.out.println("DEBUG: this.uncovered_sizes = " + this.uncovered_sizes);
	e.printStackTrace();
}
  }
  
  private Map<Integer, Integer> createUncoveredSizeArray(Integer itemNr) {
    Map<Integer, Integer> new_size = new TreeMap<Integer, Integer> ();
    
    Set<Integer> tidlist = this.uncovered_tidlists.get(itemNr);
    
    Iterator<Integer> transactions = this.uncovered_sizes.keySet().iterator();
    while (transactions.hasNext()) {
      Integer transaction = transactions.next();
      Integer size = this.uncovered_sizes.get(transaction); // get the current size
      
      if (tidlist.contains(transaction)) { // decrease by 1
      	size--;
      }
			assert size >= 0;
//      if (size > 0) { // no need to keep a size for a transaction that doesn't occur in a tidlist
      	new_size.put(transaction, size);
//      }
    }
    
    return new_size;
  }
  
  /**
   * Create D^i: conditional database over i
   * @param i one of the items
   * @return the tidlists of the conditional database
   */
  private Map<Integer, Set<Integer>> conditional_uncovered_tidlists (Integer i) {
    Map<Integer, Set<Integer>> conditional_i = new TreeMap<Integer, Set<Integer>> ();
    Set<Integer> tidlist_i = this.tidlists.get(i);
    
    Set<Integer> copy = new TreeSet<Integer> (this.uncovered_tidlists.keySet());
    Iterator<Integer> keys = copy.iterator();
    
    Integer j = keys.next();
    while (j != i) { // loop till (j == i)
      j = keys.next();
    }
    // for every j > i, make the intersection of the tidlists of i and j, then add it to the database
    while (keys.hasNext()) {
      j = keys.next();
      Set<Integer> intersection = Utility.intersectSets(tidlist_i, this.uncovered_tidlists.get(j));
      conditional_i.put(j, intersection);
    }
    
    return conditional_i;
  }
  
  /**
   * Create a database that is conditional over a given item
   * @param itemNr
   * @return the conditional database
   */
  public LTMDatabase conditional_database (Integer itemNr) {
    Map<Integer, Set<Integer>> new_tidlists = this.conditional_tidlists(itemNr);
    Map<Integer, Set<Integer>> new_uncovered_tidlists = this.conditional_uncovered_tidlists(itemNr);
    if (new_tidlists.size() > 0) {
      Map<Integer, Integer> new_sizes = this.createSizeArray(itemNr);
      Map<Integer, Integer> new_uncovered_sizes = this.createUncoveredSizeArray(itemNr);
      
      Set<Integer> new_itemset = new TreeSet<Integer> (this.itemset);
      new_itemset.add(itemNr);
      
      Map<Integer, Set<Integer>> new_uncovered_items = new TreeMap<Integer, Set<Integer>>(this.uncovered_items);
      new_uncovered_items.put(itemNr, new TreeSet<Integer>(this.uncovered_tidlists.get(itemNr)));
      
      TilingOverlapDatabase new_db = new TilingOverlapDatabase(new_tidlists, new_sizes, new_itemset,
      		new_uncovered_tidlists, new_uncovered_sizes, new_uncovered_items);
      return new_db;
    }
    else {
      return null;
    }
  }
  
  /**
   * Removes a transaction from the tidlist of a given item
   * @param itemNr
   * @param tid
   */
  public void removeTransactionFromTidlist (Integer itemNr, Integer tid) {
    this.tidlists.get(itemNr).remove(tid);
    this.uncovered_tidlists.get(itemNr).remove(tid);
    this.lowerSize(tid);
  }
  
  /**
   * Calls the top-k algorithm for large tiles
   * @param k
   * @return the k largest tiles of the database
   */
  public Vector<Set<Integer>> topK (int k, Vector<Integer> sizes) {
    Vector<Set<Integer>> tiles = new Vector<Set<Integer>> (k);
    tiles.add(new TreeSet<Integer> ());
    //Vector<Integer> sizes = new Vector<Integer> (k);
    sizes.add(0);
    this.LTMTopK(tiles, sizes);
    return tiles;
  }
  
  /**
   * Calls the LTM algorithm 
   * @param tiles list of tiles found so far
   * @param sizes list of sizes for the tiles found so far
   */
  public void LTMTopK (Vector<Set<Integer>> tiles, Vector<Integer> sizes) {
    // prune the database
    this.prune(sizes.get(0));
    
    Iterator<Integer> keys = this.tidlists.keySet().iterator();
    while (keys.hasNext()) { // loop for every item in the database
      
      Integer i = keys.next();
      
      int my_size = 0;
      Set<Integer> tidlist = this.tidlists.get(i);
      Iterator<Integer> it = this.uncovered_items.keySet().iterator();
      while (it.hasNext()) { // loop for every item in the itemset
      	my_size += Utility.intersectSets(this.uncovered_items.get(it.next()), tidlist).size();
      }
      my_size += this.uncovered_tidlists.get(i).size();
      
      // If a large tile, add the set to the set of large tiles
      if (my_size > sizes.get(0)) {
      	// copy the itemset and add i
      	Set<Integer> itemset_i = new TreeSet<Integer> (this.itemset);
      	itemset_i.add(i);
      	
      	this.insertTile(tiles, sizes, itemset_i, my_size);
      }
      
      //Create D^i: conditional database over i
      LTMDatabase conditional_i = this.conditional_database(i);
      
      // if conditional database not empty, apply recursively
      if ((conditional_i != null) && (my_size > 0)) {
      	conditional_i.LTMTopK(tiles, sizes);
      }
    }
  }
  
  public void insertTile (Vector<Set<Integer>> tiles, Vector<Integer> sizes, Set<Integer> my_tile, int my_size) {
    if (tiles.size() == tiles.capacity()) {
      tiles.remove(0);
      sizes.remove(0);
    }
    for (int index = 0; index < tiles.size(); index++) {
      if (my_size <= sizes.get(index)) {
      	tiles.insertElementAt(my_tile, index);
      	sizes.insertElementAt(my_size, index);
      	return;
      }
    }
    tiles.add(my_tile);
    sizes.add(my_size);
  }
  
  /**
   * Function to prune the database
   * @param minTiling the minimum size of a large tile
   */
  public void prune (int minTiling) {
    boolean changed; // check whether something has changed while pruning
    do {
      changed = false;
      Set<Integer> copy = new TreeSet<Integer> (this.tidlists.keySet());
      Iterator<Integer> keys = copy.iterator();
      while (keys.hasNext()) { // loop for every item in the database
      	Integer i = keys.next();
      	
      	Map<Integer, Integer> counts = new TreeMap<Integer, Integer> ();
      	int maximum = this.countTransactionSizes(counts, i);
      	
      	// check whether it's possible to create a large tile with this item
      	boolean possible = false;
      	int amount = 0; // amount counts all the transactions with a size which is >= than the current size
      	int ML = maximum;
      	for (int size = maximum; size > 0; size--) { // start with the highest sizes, then count down to the lowest
          if (counts.containsKey(size)) {
            amount += counts.get(size);
            int upper_bound = amount * (this.itemset.size() + size); // the maximum area that can be covered with this amount and size
            if (upper_bound >= minTiling) {
              possible = true;
              ML = size; // we want to keep the minimum size for which a large tile is still possible
            }
          }
      	}
      	
      	int UB2 = 0;
      	Set<Integer> tidlist = this.tidlists.get(i);
      	Iterator<Integer> it = this.uncovered_items.keySet().iterator();
      	while (it.hasNext()) {
      	  UB2 += Utility.intersectSets(this.uncovered_items.get(it.next()), tidlist).size();
      	}
      	
      	it = this.tidlists.get(i).iterator();
      	while (it.hasNext()) {
          int key = it.next();
          if (this.uncovered_sizes.get(key) != null) {
            UB2 += this.uncovered_sizes.get(key);
          }
      	}
		
				// BUG: UB2 not used
      	
      	if (!possible) { // if not possible, remove the item
      	  changed = true;
      	  this.removeItem(i);
      	}
      	else { // it's possible to create a large tile, but maybe not for all transactions
          
          Set<Integer> tidlist_i = this.tidlists.get(i);
          copy = new TreeSet<Integer> (tidlist_i);
          Iterator<Integer> transactions = copy.iterator();
          
          while (transactions.hasNext()) {
            Integer transaction = transactions.next();
            if (this.sizes.get(transaction) < ML) { // remove if size < minimal size needed
              this.removeTransactionFromTidlist(i, transaction);
              changed = true;
            }
          }
      	}
      }
    } while (changed); // loop till nothing changed
  }
  
  public FastVector[] CallTiling (int k) {
    Vector<Integer> sizes = new Vector<Integer> (k);
    Vector<Set<Integer>> result = this.mine(k, sizes);
    FastVector[] copy = new FastVector[result.size()+1];
    
    FastVector copySizes = new FastVector(sizes.size());
    for (int i = 0; i < sizes.size(); i++) {
      copySizes.addElement(sizes.get(i));
    }
    
    copy[0] = copySizes;
    
    for (int i = 0; i < result.size(); i++) {
      copy[i+1] = Utility.SetToFastVector(result.get(i));
    }
    
    return copy;
  }
  
  public Vector<Set<Integer>> mine(int k, Vector<Integer> sizes) {
    Vector<Set<Integer>> tiling = new Vector<Set<Integer>>();
    for (int i = 0; i < k; i++) {
      Vector<Integer> size = new Vector<Integer> (1);
      Set<Integer> largest = this.topK(1, size).firstElement();
      if (largest.size() > 0) {
      	tiling.add(largest);
      	sizes.add(size.get(0));
System.out.print(new Time(System.currentTimeMillis()) + ", additional size: " + size.get(0));
System.out.println(largest);
      } else {
      	break;
      }
      this.DeleteTile(largest);
    }
    
    return tiling;
  }
}
